from astris.component import Element, Text


class Div(Element):
    tag = "div"


def test_text_render_and_str() -> None:
    text = Text("hola")

    assert text.render() == "hola"
    assert str(text) == "hola"


def test_element_transforms_html_attributes() -> None:
    element = Div(class_name="hero", data_test_id="banner")

    assert element.attributes["class"] == "hero"
    assert element.attributes["data-test-id"] == "banner"


def test_element_renders_children_components_and_strings() -> None:
    element = Div(children=[Text("A"), "B", Div(children=["C"])])

    assert element.render() == "<div>A B<div>C</div></div>".replace(" ", "")



import os

from astris import AstrisApp


def test_run_dev_uses_reload_import_string_when_available(monkeypatch) -> None:
    app = AstrisApp()
    calls = []

    monkeypatch.setattr(app, "_infer_import_string", lambda: "example:app._fastapi_app")

    def fake_run(*args, **kwargs):
        calls.append((args, kwargs))

    monkeypatch.setattr("astris.app.uvicorn.run", fake_run)

    app.run_dev(port=9001, reload=True)

    assert len(calls) == 1
    args, kwargs = calls[0]
    assert args[0] == "example:app._fastapi_app"
    assert kwargs["reload"] is True
    assert kwargs["port"] == 9001
    assert kwargs["reload_dirs"] == [os.getcwd()]


def test_run_dev_falls_back_when_import_string_is_missing(monkeypatch) -> None:
    app = AstrisApp()
    calls = []

    monkeypatch.setattr(app, "_infer_import_string", lambda: None)

    def fake_run(*args, **kwargs):
        calls.append((args, kwargs))

    monkeypatch.setattr("astris.app.uvicorn.run", fake_run)

    app.run_dev(port=9002, reload=True)

    assert len(calls) == 1
    args, kwargs = calls[0]
    assert args[0] is app._fastapi_app
    assert kwargs["port"] == 9002
    assert "reload" not in kwargs


def test_run_dev_without_reload_uses_fastapi_app(monkeypatch) -> None:
    app = AstrisApp()
    calls = []

    def fake_run(*args, **kwargs):
        calls.append((args, kwargs))

    monkeypatch.setattr("astris.app.uvicorn.run", fake_run)

    app.run_dev(port=9003, reload=False)

    assert len(calls) == 1
    args, kwargs = calls[0]
    assert args[0] is app._fastapi_app
    assert kwargs["port"] == 9003



from __future__ import annotations

import argparse
import importlib.util
import re
import sys
from pathlib import Path

from .app import AstrisApp

MAIN_TEMPLATE = """import sys

from astris import AstrisApp
from astris.lib import Body, H1, Html

app = AstrisApp()


@app.page("/")
def home():
    return Html(children=[
        Body(children=[
            H1(children=["Hello, world from Astris!"]),
        ])
    ])


if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == "build":
        app.build()
    else:
        app.run_dev()
"""

README_TEMPLATE = """# {project_name}

Project generated with `astris new`.

## Run in development

```bash
uv run python main.py
```

## Build static site

```bash
uv run astris build
```
"""

GITIGNORE_TEMPLATE = """__pycache__/
*.pyc
.venv/
dist/
"""

REQUIREMENTS_TEMPLATE = "astris\n"


class CliError(ValueError):
    """CLI usage error for project generation."""


def validate_project_name(project_name: str) -> None:
    """Validate that the provided project name is filesystem friendly."""
    pattern = r"^[A-Za-z][A-Za-z0-9_-]*$"
    if not re.match(pattern, project_name):
        raise CliError(
            "Invalid project name. Use letters, numbers, '_' or '-', and start with a letter."
        )


def create_project(project_name: str, base_path: Path | None = None) -> Path:
    """Create a minimal Astris project scaffold."""
    validate_project_name(project_name)

    root = (base_path or Path.cwd()) / project_name
    if root.exists():
        raise CliError(f"Target directory already exists: {root}")

    root.mkdir(parents=True, exist_ok=False)

    (root / "main.py").write_text(MAIN_TEMPLATE, encoding="utf-8")
    (root / "README.md").write_text(
        README_TEMPLATE.format(project_name=project_name), encoding="utf-8"
    )
    (root / ".gitignore").write_text(GITIGNORE_TEMPLATE, encoding="utf-8")
    (root / "requirements.txt").write_text(REQUIREMENTS_TEMPLATE, encoding="utf-8")

    return root


def _load_app_from_file(file_path: str) -> AstrisApp:
    target = Path(file_path).expanduser()
    if not target.is_absolute():
        target = Path.cwd() / target
    target = target.resolve()

    if not target.exists():
        raise CliError(f"Python file not found: {target}")
    if not target.is_file():
        raise CliError(f"Path is not a file: {target}")

    module_name = f"astris_build_target_{target.stem}"
    spec = importlib.util.spec_from_file_location(module_name, target)
    if spec is None or spec.loader is None:
        raise CliError(f"Could not load module from: {target}")

    module = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(module)
    except Exception as error:  # pragma: no cover - exercised via CLI behavior
        raise CliError(f"Failed to execute {target}: {error}") from error

    app = getattr(module, "app", None)
    if app is None:
        raise CliError(f"No 'app' variable found in {target}")
    if not isinstance(app, AstrisApp):
        raise CliError(f"'app' in {target} must be an AstrisApp instance")

    return app


def build_parser() -> argparse.ArgumentParser:
    """Build the top-level Astris CLI parser."""
    parser = argparse.ArgumentParser(
        prog="astris",
        description="Astris CLI",
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    new_parser = subparsers.add_parser("new", help="Create a new Astris project")
    new_parser.add_argument("project_name", help="Name of the project directory")

    build_parser = subparsers.add_parser("build", help="Build static site")
    build_parser.add_argument(
        "--file",
        default="main.py",
        help="Python app file to load (default: main.py)",
    )

    return parser


def main(argv: list[str] | None = None) -> int:
    """CLI entrypoint."""
    parser = build_parser()
    args = parser.parse_args(argv)

    if args.command == "new":
        try:
            created = create_project(args.project_name)
        except CliError as error:
            print(f"Error: {error}", file=sys.stderr)
            return 2

        print(f"Created Astris project at {created}")
        print("Next steps:")
        print(f"  cd {args.project_name}")
        print("  uv run python main.py")
        return 0

    if args.command == "build":
        try:
            app = _load_app_from_file(args.file)
            app.build()
        except CliError as error:
            print(f"Error: {error}", file=sys.stderr)
            return 2

        print("Build complete: dist/")
        return 0

    parser.print_help()
    return 1


if __name__ == "__main__":
    raise SystemExit(main())



from pathlib import Path

from astris import AstrisApp, Text
from astris.lib import A, Div


def test_build_generates_static_files_and_doctype(tmp_path: Path) -> None:
    app = AstrisApp()

    @app.page("/")
    def home():
        return Div(
            children=[
                A(href="/about", children=["About"]),
                Text(" "),
                A(href="https://example.com", children=["Externo"]),
            ]
        )

    @app.page("/about")
    def about():
        return Div(
            children=[
                A(href="/", children=["Inicio"]),
                A(href="/about?tab=team#top", children=["Anchor"]),
            ]
        )

    output_dir = tmp_path / "site"
    app.build(str(output_dir))

    index_file = output_dir / "index.html"
    about_file = output_dir / "about.html"

    assert index_file.exists()
    assert about_file.exists()

    index_html = index_file.read_text(encoding="utf-8")
    about_html = about_file.read_text(encoding="utf-8")

    assert index_html.startswith("<!DOCTYPE html>")
    assert about_html.startswith("<!DOCTYPE html>")
    assert 'href="about.html"' in index_html
    assert 'href="https://example.com"' in index_html
    assert 'href="index.html"' in about_html
    assert 'href="about.html?tab=team#top"' in about_html


def test_build_rewrites_relative_links_for_nested_routes(tmp_path: Path) -> None:
    app = AstrisApp()

    @app.page("/about")
    def about():
        return Div(children=[Text("About")])

    @app.page("/docs/getting-started")
    def docs():
        return Div(
            children=[
                A(href="/about", children=["About"]),
                A(href="/missing", children=["Missing"]),
            ]
        )

    output_dir = tmp_path / "site"
    app.build(str(output_dir))

    docs_html = (output_dir / "docs/getting-started.html").read_text(encoding="utf-8")

    assert 'href="../about.html"' in docs_html
    assert 'href="/missing"' in docs_html


def test_build_includes_registered_head_assets(tmp_path: Path) -> None:
    app = AstrisApp()
    app.add_head_link("https://cdn.jsdelivr.net/npm/bootstrap@5.3.0/dist/css/bootstrap.min.css")
    app.add_head_script("https://cdn.jsdelivr.net/npm/bootstrap@5.3.0/dist/js/bootstrap.bundle.min.js")

    @app.page("/")
    def home():
        return Div(
            children=[
                Text(
                    "<html><head><title>Home</title></head><body><h1>Hi</h1></body></html>"
                )
            ]
        )

    output_dir = tmp_path / "site"
    app.build(str(output_dir))

    index_html = (output_dir / "index.html").read_text(encoding="utf-8")

    assert "bootstrap.min.css" in index_html
    assert "bootstrap.bundle.min.js" in index_html



from pathlib import Path

from astris import cli


def test_create_project_generates_expected_files(tmp_path: Path) -> None:
    project_path = cli.create_project("demo_site", base_path=tmp_path)

    assert project_path == tmp_path / "demo_site"
    assert (project_path / "main.py").exists()
    assert (project_path / "README.md").exists()
    assert (project_path / ".gitignore").exists()
    assert (project_path / "requirements.txt").exists()

    main_content = (project_path / "main.py").read_text(encoding="utf-8")
    assert "app = AstrisApp()" in main_content
    assert '@app.page("/")' in main_content


def test_create_project_fails_if_target_exists(tmp_path: Path) -> None:
    existing = tmp_path / "demo_site"
    existing.mkdir()

    try:
        cli.create_project("demo_site", base_path=tmp_path)
        assert False, "Expected CliError"
    except cli.CliError as error:
        assert "already exists" in str(error)


def test_create_project_rejects_invalid_name(tmp_path: Path) -> None:
    try:
        cli.create_project("123-invalid", base_path=tmp_path)
        assert False, "Expected CliError"
    except cli.CliError as error:
        assert "Invalid project name" in str(error)


def test_main_new_command_returns_success(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    exit_code = cli.main(["new", "hello_app"])

    assert exit_code == 0
    assert (tmp_path / "hello_app" / "main.py").exists()


def test_main_build_command_uses_main_py_by_default(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "main.py").write_text(
        """
from astris import AstrisApp
from astris.lib import Body, Html

app = AstrisApp()


@app.page("/")
def home():
    return Html(children=[Body(children=["hello"])])
""".strip(),
        encoding="utf-8",
    )

    exit_code = cli.main(["build"])

    assert exit_code == 0
    assert (tmp_path / "dist" / "index.html").exists()


def test_main_build_command_supports_custom_file(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "example.py").write_text(
        """
from astris import AstrisApp
from astris.lib import Body, Html

app = AstrisApp()


@app.page("/")
def home():
    return Html(children=[Body(children=["hello from example"])])
""".strip(),
        encoding="utf-8",
    )

    exit_code = cli.main(["build", "--file", "example.py"])

    assert exit_code == 0
    output = (tmp_path / "dist" / "index.html").read_text(encoding="utf-8")
    assert "hello from example" in output


def test_main_build_command_fails_if_file_missing(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    exit_code = cli.main(["build"])

    assert exit_code == 2


def test_main_build_command_fails_if_app_not_found(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "main.py").write_text("x = 1", encoding="utf-8")

    exit_code = cli.main(["build"])

    assert exit_code == 2


def test_main_build_command_fails_if_app_type_is_invalid(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "main.py").write_text("app = object()", encoding="utf-8")

    exit_code = cli.main(["build"])

    assert exit_code == 2



from astris.lib import A, Column, Container, Head, Html, Row, Title


def test_html_wrappers_render_expected_tags() -> None:
    page = Html(
        children=[
            Head(children=[Title(children=["Demo"])]),
            Container(children=[A(href="/", children=["Inicio"])]),
        ]
    )
    html = page.render()

    assert "<html>" in html
    assert "<head>" in html
    assert "<title>Demo</title>" in html
    assert '<a href="/">Inicio</a>' in html


def test_column_and_row_include_default_flex_style() -> None:
    column = Column(children=["x"])
    row = Row(children=["y"])

    assert "flex-direction: column" in column.render()
    assert "flex-direction: row" in row.render()



from .component import Element


class Html(Element):
    tag = "html"


class Head(Element):
    tag = "head"


class Body(Element):
    tag = "body"


class Div(Element):
    tag = "div"


class Span(Element):
    tag = "span"


class H1(Element):
    tag = "h1"


class H2(Element):
    tag = "h2"


class P(Element):
    tag = "p"


class A(Element):
    tag = "a"


class Ul(Element):
    tag = "ul"


class Li(Element):
    tag = "li"


class Button(Element):
    tag = "button"


class Title(Element):
    tag = "title"


class Container(Div):
    pass


class Column(Div):
    def __init__(self, children=None, **kwargs):
        super().__init__(
            children, style="display: flex; flex-direction: column;", **kwargs
        )


class Row(Div):
    def __init__(self, children=None, **kwargs):
        super().__init__(
            children, style="display: flex; flex-direction: row;", **kwargs
        )



import os
import re
import sys
from pathlib import Path
from typing import Dict, List
from urllib.parse import urlsplit, urlunsplit

import uvicorn
from fastapi import FastAPI
from fastapi.responses import HTMLResponse

from .component import Component


class AstrisApp:
    def __init__(self):
        self.routes: Dict[str, Component] = {}
        self._head_links: List[Dict[str, str]] = []
        self._head_scripts: List[Dict[str, str]] = []
        self._fastapi_app = FastAPI()

    def add_head_link(self, href: str, rel: str = "stylesheet", **attributes) -> None:
        """Register a <link> tag to be injected in the document head."""
        link_attrs = {"rel": rel, "href": href, **attributes}
        self._head_links.append({key: str(value) for key, value in link_attrs.items()})

    def add_head_script(self, src: str, **attributes) -> None:
        """Register a <script> tag to be injected in the document head."""
        script_attrs = {"src": src, **attributes}
        self._head_scripts.append(
            {key: str(value) for key, value in script_attrs.items()}
        )

    def _render_tag_attributes(self, attributes: Dict[str, str]) -> str:
        return " ".join([f'{key}="{value}"' for key, value in attributes.items()])

    def _render_head_assets(self) -> str:
        parts: List[str] = []
        for link_attrs in self._head_links:
            attrs = self._render_tag_attributes(link_attrs)
            parts.append(f"<link {attrs}>")

        for script_attrs in self._head_scripts:
            attrs = self._render_tag_attributes(script_attrs)
            parts.append(f"<script {attrs}></script>")

        return "".join(parts)

    def _inject_head_assets(self, html: str) -> str:
        head_assets = self._render_head_assets()
        if not head_assets:
            return html

        head_close_match = re.search(r"</head>", html, re.IGNORECASE)
        if head_close_match:
            insert_at = head_close_match.start()
            return f"{html[:insert_at]}{head_assets}{html[insert_at:]}"

        html_open_match = re.search(r"<html[^>]*>", html, re.IGNORECASE)
        if html_open_match:
            insert_at = html_open_match.end()
            return f"{html[:insert_at]}<head>{head_assets}</head>{html[insert_at:]}"

        return f"<head>{head_assets}</head>{html}"

    def _render_page_html(self, component: Component) -> str:
        return self._inject_head_assets(component.render())

    def page(self, path: str):
        """Decorator to register a page (route)."""

        def decorator(func):
            component_tree = func()
            self.routes[path] = component_tree

            @self._fastapi_app.get(path, response_class=HTMLResponse)
            async def serve_page():
                return f"<!DOCTYPE html>{self._render_page_html(component_tree)}"

            return func

        return decorator

    def _infer_import_string(self) -> str | None:
        main_module = sys.modules.get("__main__")
        main_file = getattr(main_module, "__file__", None)
        if not main_file:
            return None

        module_name = Path(main_file).stem
        return f"{module_name}:app._fastapi_app"

    def run_dev(self, port=8000, reload=True):
        """Start the development server."""
        print(f"🚀 Dev server running at http://localhost:{port}")

        if reload:
            import_string = self._infer_import_string()
            if import_string:
                uvicorn.run(
                    import_string,
                    host="0.0.0.0",
                    port=port,
                    reload=True,
                    reload_dirs=[os.getcwd()],
                )
                return

            print("⚠️ Could not infer the main module, starting without hot reload.")

        uvicorn.run(self._fastapi_app, host="0.0.0.0", port=port)

    def _route_to_filename(self, route: str) -> str:
        normalized = route if route.startswith("/") else f"/{route}"
        if normalized == "/":
            return "index.html"
        return f"{normalized.strip('/')}.html"

    def _resolve_route(self, path: str) -> str | None:
        if path in self.routes:
            return path

        without_slash = path.rstrip("/")
        if without_slash and without_slash in self.routes:
            return without_slash

        with_slash = f"{without_slash}/"
        if with_slash in self.routes:
            return with_slash

        return None

    def _rewrite_static_links(self, current_route: str, html: str) -> str:
        current_file = self._route_to_filename(current_route)
        current_dir = os.path.dirname(current_file) or "."

        def replace_href(match: re.Match[str]) -> str:
            quote = match.group("quote")
            href_value = match.group("href")

            split = urlsplit(href_value)
            if split.scheme or split.netloc or split.path.startswith("//"):
                return match.group(0)

            resolved_route = self._resolve_route(split.path)
            if not resolved_route:
                return match.group(0)

            target_file = self._route_to_filename(resolved_route)
            relative_target = os.path.relpath(target_file, start=current_dir).replace(
                os.sep, "/"
            )
            rebuilt_href = urlunsplit(
                ("", "", relative_target, split.query, split.fragment)
            )

            return f"href={quote}{rebuilt_href}{quote}"

        return re.sub(
            r'href=(?P<quote>["\'])(?P<href>.*?)(?P=quote)',
            replace_href,
            html,
        )

    def build(self, output_dir="dist"):
        """Generate static HTML files."""
        print(f"📦 Building site into ./{output_dir}...")

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        for path, component in self.routes.items():
            filename = self._route_to_filename(path)
            filepath = os.path.join(output_dir, filename)
            os.makedirs(os.path.dirname(filepath), exist_ok=True)

            rendered_html = self._rewrite_static_links(
                path, self._render_page_html(component)
            )

            with open(filepath, "w", encoding="utf-8") as f:
                f.write("<!DOCTYPE html>\n")
                f.write(rendered_html)

            print(f"  ✅ Generated: {filepath}")

        print("✨ Build completed successfully.")



import types

from astris import AstrisApp, Text
from astris.lib import Div


def test_page_decorator_registers_route() -> None:
    app = AstrisApp()

    @app.page("/")
    def home():
        return Div(children=[Text("home")])

    assert "/" in app.routes
    assert app.routes["/"].render() == "<div>home</div>"


def test_route_helpers() -> None:
    app = AstrisApp()
    app.routes = {"/": Div(), "/about": Div(), "/blog/": Div()}

    assert app._route_to_filename("/") == "index.html"
    assert app._route_to_filename("/about") == "about.html"
    assert app._route_to_filename("docs/setup") == "docs/setup.html"

    assert app._resolve_route("/about") == "/about"
    assert app._resolve_route("/blog") == "/blog/"
    assert app._resolve_route("/missing") is None


def test_infer_import_string_from_main_module(monkeypatch) -> None:
    app = AstrisApp()
    fake_main = types.SimpleNamespace(__file__="/tmp/example.py")
    monkeypatch.setitem(__import__("sys").modules, "__main__", fake_main)

    assert app._infer_import_string() == "example:app._fastapi_app"


def test_render_page_html_injects_registered_head_assets() -> None:
    app = AstrisApp()
    app.add_head_link("https://cdn.example.com/bootstrap.css")
    app.add_head_script("https://cdn.example.com/app.js", defer="defer")

    html = app._render_page_html(
        Div(
            children=[
                Text("<html><head><title>Home</title></head><body>Hello</body></html>")
            ]
        )
    )

    assert (
        '<link rel="stylesheet" href="https://cdn.example.com/bootstrap.css">' in html
    )
    assert '<script src="https://cdn.example.com/app.js" defer="defer"></script>' in html
    assert html.index("bootstrap.css") < html.index("</head>")


def test_render_page_html_creates_head_when_missing() -> None:
    app = AstrisApp()
    app.add_head_link("https://cdn.example.com/tailwind.css")

    html = app._render_page_html(Div(children=[Text("<html><body>Hello</body></html>")]))

    assert "<head>" in html
    assert "tailwind.css" in html



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Sequence, Union


class Component(ABC):
    """Base class for all UI elements (widgets)."""

    @abstractmethod
    def render(self) -> str:
        pass

    def __str__(self):
        return self.render()


class Text(Component):
    """Render plain text without tags."""

    def __init__(self, content: str):
        self.content = content

    def render(self) -> str:
        return str(self.content)


class Element(Component):
    """
    Represents a generic HTML tag.
    Works like a Flutter-style container that accepts children.
    """

    tag: str = "div"

    def __init__(
        self, children: Optional[Sequence[Union[Component, str]]] = None, **attributes
    ):
        self.children: List[Union[Component, str]] = list(children) if children else []
        self.attributes = self._process_attributes(attributes)

    def _process_attributes(self, attrs: Dict) -> Dict:
        processed = {}
        for key, value in attrs.items():
            normalized_key = key.replace("class_name", "class").replace("_", "-")
            processed[normalized_key] = value
        return processed

    def render(self) -> str:
        attrs_str = " ".join([f'{k}="{v}"' for k, v in self.attributes.items()])
        attrs_str = f" {attrs_str}" if attrs_str else ""

        children_html = ""
        for child in self.children:
            if isinstance(child, Component):
                children_html += child.render()
            else:
                children_html += str(child)

        return f"<{self.tag}{attrs_str}>{children_html}</{self.tag}>"
